/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2012 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_main.c  -- client main loop

#include "quakedef.h"

// we need to declare some mouse variables here, because the menu system
// references them even when on a unix system.

// these two are not intended to be set directly
cvar_t	cl_name = {"_cl_name", "player", true};
cvar_t	cl_color = {"_cl_color", "0", true};

cvar_t	cl_shownet = {"cl_shownet","0"};	// can be 0, 1, or 2
cvar_t	cl_nolerp = {"cl_nolerp","0"};

cvar_t	lookspring = {"lookspring","0", true};
cvar_t	lookstrafe = {"lookstrafe","0", true};
cvar_t	sensitivity = {"sensitivity","3", true};

cvar_t	m_pitch = {"m_pitch","0.022", true};
cvar_t	m_yaw = {"m_yaw","0.022", true};
cvar_t	m_forward = {"m_forward","1", true};
cvar_t	m_side = {"m_side","0.8", true};

cvar_t	cl_maxpitch = {"cl_maxpitch", "90", true}; //johnfitz -- variable pitch clamping
cvar_t	cl_minpitch = {"cl_minpitch", "-90", true}; //johnfitz -- variable pitch clamping

client_static_t	cls;
client_state_t	cl;
// FIXME: put these on hunk?
efrag_t			cl_efrags[MAX_EFRAGS];
entity_t		cl_static_entities[MAX_STATIC_ENTITIES];
lightstyle_t	cl_lightstyle[MAX_LIGHTSTYLES];
dlight_t		cl_dlights[MAX_DLIGHTS];

entity_t		*cl_entities; //johnfitz -- was a static array, now on hunk
int				cl_max_edicts; //johnfitz -- only changes when new map loads

int				cl_numvisedicts;
entity_t		*cl_visedicts[MAX_VISEDICTS];

extern cvar_t	r_lerpmodels, r_lerpmove; //johnfitz

/*
=====================
CL_ClearState

=====================
*/
void CL_ClearState (void)
{
	int			i;

	if (!sv.active)
		Host_ClearMemory ();

// wipe the entire cl structure
	memset (&cl, 0, sizeof(cl));

	SZ_Clear (&cls.message);

// clear other arrays
	memset (cl_efrags, 0, sizeof(cl_efrags));
	memset (cl_dlights, 0, sizeof(cl_dlights));
	memset (cl_lightstyle, 0, sizeof(cl_lightstyle));
	memset (cl_temp_entities, 0, sizeof(cl_temp_entities));
	memset (cl_beams, 0, sizeof(cl_beams));

	//johnfitz -- cl_entities is now dynamically allocated
#ifdef SUPPORTS_SMART_EDICTS // Baker change

	if (sv.active)
		cl_max_edicts = sv.max_edicts; // Server is active
	else if (max_edicts.value)
		cl_max_edicts = max_edicts.value; // Client-only (demo or connected to a server) and max_edicts has non-zero value (should be rare)
	else
		cl_max_edicts = 8192; // Client-only and max_edicts is 0.  Set to 8192.  cl edicts are tiny.

	cl_max_edicts = CLAMP (MIN_EDICTS, cl_max_edicts, MAX_EDICTS);
#else // Baker change +
	cl_max_edicts = CLAMP (MIN_EDICTS,(int)max_edicts.value,MAX_EDICTS);
	//johnfitz
#endif
	cl_entities = Hunk_AllocName (cl_max_edicts*sizeof(entity_t), "cl_entities");

//
// allocate the efrags and chain together into a free list
//
	cl.free_efrags = cl_efrags;
	for (i=0 ; i<MAX_EFRAGS-1 ; i++)
		cl.free_efrags[i].entnext = &cl.free_efrags[i+1];
	cl.free_efrags[i].entnext = NULL;
}

/*
=====================
CL_Disconnect

Sends a disconnect message to the server
This is also called on Host_Error, so it shouldn't cause any errors
=====================
*/
void CL_Disconnect (void)
{
// stop sounds (especially looping!)
	S_StopAllSounds (true);

#ifdef GENERIC_TOUCH_UP // Baker change
	CDAudio_Stop(); // Stop CD music
#endif // Baker change +

// if running a local server, shut it down
	if (cls.demoplayback)
		CL_StopPlayback ();
	else if (cls.state == ca_connected)
	{
		if (cls.demorecording)
			CL_Stop_f ();

		Con_DPrintf ("Sending clc_disconnect\n");
		SZ_Clear (&cls.message);
		MSG_WriteByte (&cls.message, clc_disconnect);
		NET_SendUnreliableMessage (cls.netcon, &cls.message);
		SZ_Clear (&cls.message);
		NET_Close (cls.netcon);

		cls.state = ca_disconnected;
		if (sv.active)
			Host_ShutdownServer(false);
	}

	cls.demoplayback = cls.timedemo = false;
	cls.signon = 0;
#ifdef SUPPORTS_DEMO_END_AT_INTERMISSION // Baker change
	cl.intermission = 0; // Baker: So critical.  SCR_UpdateScreen uses this.
//	SCR_EndLoadingPlaque ();  // Baker: NOOOOOO.  This shows between start demos.  We need this.
#endif // Baker change +
}

void CL_Disconnect_f (void)
{
#ifdef SUPPORTS_CLEAR_DEMOQUEUE // Baker change
	CL_Clear_Demos_Queue (); // timedemo is a very intentional action
#endif // Baker change +

	CL_Disconnect ();
	if (sv.active)
		Host_ShutdownServer (false);
}




/*
=====================
CL_EstablishConnection

Host should be either "local" or a net address to be passed on
=====================
*/
void CL_EstablishConnection (char *host)
{
	if (cls.state == ca_dedicated)
		return;

	if (cls.demoplayback)
		return;

	CL_Disconnect ();

	cls.netcon = NET_Connect (host);
	if (!cls.netcon)
		Host_Error ("CL_Connect: connect failed\n");
	Con_DPrintf ("CL_EstablishConnection: connected to %s\n", host);

	cls.demonum = -1;			// not in the demo loop now
	cls.state = ca_connected;
	cls.signon = 0;				// need all the signon messages before playing

#ifdef SUPPORTS_EASY_NAT_FIX // Baker change
	MSG_WriteByte (&cls.message, clc_nop); // NAT fix.  Covers many, but not all, circumstances.
#endif // Baker change +
}

/*
=====================
CL_SignonReply

An svc_signonnum has been received, perform a client side setup
=====================
*/
void CL_SignonReply (void)
{
	char 	str[8192];

Con_DPrintf ("CL_SignonReply: %i\n", cls.signon);

	switch (cls.signon)
	{
	case 1:
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "prespawn");
		break;

	case 2:
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("name \"%s\"\n", cl_name.string));

		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("color %i %i\n", ((int)cl_color.value)>>4, ((int)cl_color.value)&15));

		MSG_WriteByte (&cls.message, clc_stringcmd);
		sprintf (str, "spawn %s", cls.spawnparms);
		MSG_WriteString (&cls.message, str);
		break;

	case 3:
		MSG_WriteByte (&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "begin");
		Cache_Report ();		// print remaining memory
		break;

	case 4:
		SCR_EndLoadingPlaque ();		// allow normal screen updates
		break;
	}
}

/*
=====================
CL_NextDemo

Called to play the next demo in the demo loop
=====================
*/
void CL_NextDemo (void)
{
	char	str[1024];

	if (cls.demonum == -1)
		return;		// don't play demos

#if !defined(TOUCHUP_RMQ_QUAKESPASM_MILD_TWEAKS) // Baker change (moved below)
	SCR_BeginLoadingPlaque ();
#endif // Baker change +

	if (!cls.demos[cls.demonum][0] || cls.demonum == MAX_DEMOS)
	{
		cls.demonum = 0;
		if (!cls.demos[cls.demonum][0])
		{
			Con_Printf ("No demos listed with startdemos\n");
			cls.demonum = -1;
#ifdef TOUCHUP_RMQ_QUAKESPASM_MILD_TWEAKS // Baker change
			CL_Disconnect ();
#endif // Baker change +
			return;
		}
	}

#ifdef TOUCHUP_RMQ_QUAKESPASM_MILD_TWEAKS // Baker change (moved to AFTER we know demo will play)
	SCR_BeginLoadingPlaque ();
#endif // Baker change +

#ifdef TOUCHUP_CLEAR_DEMOQUEUE_PLAYDEMO // Baker change
	sprintf (str,"nextstartdemo %s\n", cls.demos[cls.demonum]);
#else // Baker change +
	sprintf (str,"playdemo %s\n", cls.demos[cls.demonum]);
#endif // Baker change -
	Cbuf_InsertText (str);
	cls.demonum++;
}

/*
==============
CL_PrintEntities_f
==============
*/
void CL_PrintEntities_f (void)
{
	entity_t	*ent;
	int			i;

	for (i=0,ent=cl_entities ; i<cl.num_entities ; i++,ent++)
	{
		Con_Printf ("%3i:",i);
		if (!ent->model)
		{
			Con_Printf ("EMPTY\n");
			continue;
		}
		Con_Printf ("%s:%2i  (%5.1f,%5.1f,%5.1f) [%5.1f %5.1f %5.1f]\n"
		,ent->model->name,ent->frame, ent->origin[0], ent->origin[1], ent->origin[2], ent->angles[0], ent->angles[1], ent->angles[2]);
	}
}

//johnfitz -- deleted SetPal()

/*
===============
CL_AllocDlight

===============
*/
dlight_t *CL_AllocDlight (int key)
{
	int		i;
	dlight_t	*dl;

// first look for an exact key match
	if (key)
	{
		dl = cl_dlights;
		for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
		{
			if (dl->key == key)
			{
				memset (dl, 0, sizeof(*dl));
				dl->key = key;
				dl->color[0] = dl->color[1] = dl->color[2] = 1; //johnfitz -- lit support via lordhavoc
				return dl;
			}
		}
	}

// then look for anything else
	dl = cl_dlights;
	for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
	{
		if (dl->die < cl.time)
		{
			memset (dl, 0, sizeof(*dl));
			dl->key = key;
			dl->color[0] = dl->color[1] = dl->color[2] = 1; //johnfitz -- lit support via lordhavoc
			return dl;
		}
	}

	dl = &cl_dlights[0];
	memset (dl, 0, sizeof(*dl));
	dl->key = key;
	dl->color[0] = dl->color[1] = dl->color[2] = 1; //johnfitz -- lit support via lordhavoc
	return dl;
}


/*
===============
CL_DecayLights

===============
*/
void CL_DecayLights (void)
{
	int			i;
	dlight_t	*dl;
	float		time;

#ifdef SUPPORTS_DEMO_REWIND // Baker change
	time = fabs(cl.time - cl.oldtime); // To make sure it stays forward oriented time
#else // Baker change +
	time = cl.time - cl.oldtime;
#endif // Baker change -

	dl = cl_dlights;
	for (i=0 ; i<MAX_DLIGHTS ; i++, dl++)
	{
		if (dl->die < cl.time || !dl->radius)
			continue;

		dl->radius -= time*dl->decay;
		if (dl->radius < 0)
			dl->radius = 0;
	}
}


/*
===============
CL_LerpPoint

Determines the fraction between the last two messages that the objects
should be put at.
===============
*/
float	CL_LerpPoint (void)
{
#ifdef SUPPORTS_DEMO_REWIND // Baker change
	extern qboolean bumper_on;
#endif // Baker change +
	float	f, frac;

	f = cl.mtime[0] - cl.mtime[1];

	if (!f || cls.timedemo || sv.active)
	{
#ifdef SUPPORTS_DEMO_REWIND // Baker change
		cl.time = cl.ctime = cl.mtime[0];
#else // Baker change +
		cl.time = cl.mtime[0];
#endif // Baker change -
		return 1;
	}

	if (f > 0.1) // dropped packet, or start of demo
	{
		cl.mtime[1] = cl.mtime[0] - 0.1;
		f = 0.1f ; // Baker change touch-up added f to constant (avoid MSVC Level 2+ compiler warning)
	}

#ifdef SUPPORTS_DEMO_REWIND // Baker change
	frac = (cl.ctime - cl.mtime[1]) / f;
#else // Baker change +
	frac = (cl.time - cl.mtime[1]) / f;
#endif // Baker change -

	if (frac < 0)
	{
		if (frac < -0.01)
#ifdef SUPPORTS_DEMO_REWIND // Baker change
			if (bumper_on)
			{
				cl.ctime = cl.mtime[1];
			}
			else cl.time = cl.ctime = cl.mtime[1];
#else // Baker change +
			cl.time = cl.mtime[1];
#endif // Baker change -
		frac = 0;
	}
	else if (frac > 1)
	{
		if (frac > 1.01)
#ifdef SUPPORTS_DEMO_REWIND // Baker change
			if (bumper_on)
				cl.ctime = cl.mtime[0];
			else cl.time = cl.ctime = cl.mtime[0]; // Here is where we get foobar'd
#else // Baker change +
			cl.time = cl.mtime[0];
#endif // Baker change -
		frac = 1;
	}

	//johnfitz -- better nolerp behavior
	if (cl_nolerp.value)
		return 1;
	//johnfitz	

	return frac;
}

/*
===============
CL_RelinkEntities
===============
*/
void CL_RelinkEntities (void)
{
	entity_t	*ent;
	int			i, j;
	float		frac, f, d;
	vec3_t		delta;
	float		bobjrotate;
	vec3_t		oldorg;
	dlight_t	*dl;

// determine partial update time
	frac = CL_LerpPoint ();

	cl_numvisedicts = 0;

//
// interpolate player info
//
	for (i=0 ; i<3 ; i++)
		cl.velocity[i] = cl.mvelocity[1][i] + frac * (cl.mvelocity[0][i] - cl.mvelocity[1][i]);

#ifdef SUPPORTS_VIEW_ANGLES_INTERPOLATION // Baker change
	if (cls.demoplayback || cl.last_angle_time > host_time)
#else // Baker change +
	if (cls.demoplayback)
#endif // Baker change -
	{
	// interpolate the angles
		for (j=0 ; j<3 ; j++)
		{
			d = cl.mviewangles[0][j] - cl.mviewangles[1][j];
			if (d > 180)
				d -= 360;
			else if (d < -180)
				d += 360;
#ifdef SUPPORTS_VIEW_ANGLES_INTERPOLATION // Baker change
			cl.lerpangles[j] = cl.mviewangles[1][j] + frac*d;
			if (cls.demoplayback)
#endif // Baker change +
				cl.viewangles[j] = cl.mviewangles[1][j] + frac*d;
		}
	}
#ifdef SUPPORTS_VIEW_ANGLES_INTERPOLATION // Baker change
	else
		VectorCopy(cl.viewangles, cl.lerpangles);
#endif // Baker change +

#ifdef SUPPORTS_DEMO_REWIND_HARD_ANIMS_TIME // Baker change
	bobjrotate = anglemod(100*cl.ctime);
#else // Baker change +
	bobjrotate = anglemod(100*cl.time);
#endif // Baker change -


// start on the entity after the world
	for (i=1,ent=cl_entities+1 ; i<cl.num_entities ; i++,ent++)
	{
#ifdef SUPPORTS_GHOSTING_VIA_DEMOS // Baker change
		if (!cls.ghostdemo) // Because a ghostdemo doesn't load models
#endif // Baker change +
		if (!ent->model)
		{	// empty slot
			if (ent->forcelink)
				R_RemoveEfrags (ent);	// just became empty
			continue;
		}

// if the object wasn't included in the last packet, remove it
		if (ent->msgtime != cl.mtime[0])
		{
			ent->model = NULL;
			ent->lerpflags |= LERP_RESETMOVE|LERP_RESETANIM; //johnfitz -- next time this entity slot is reused, the lerp will need to be reset
			continue;
		}

		VectorCopy (ent->origin, oldorg);

		if (ent->forcelink)
		{	// the entity was not updated in the last message
			// so move to the final spot
			VectorCopy (ent->msg_origins[0], ent->origin);
			VectorCopy (ent->msg_angles[0], ent->angles);
		}
		else
		{	// if the delta is large, assume a teleport and don't lerp
			f = frac;
			for (j=0 ; j<3 ; j++)
			{
				delta[j] = ent->msg_origins[0][j] - ent->msg_origins[1][j];
				if (delta[j] > 100 || delta[j] < -100)
				{
					f = 1;		// assume a teleportation, not a motion
					ent->lerpflags |= LERP_RESETMOVE; //johnfitz -- don't lerp teleports
				}
			}

			//johnfitz -- don't cl_lerp entities that will be r_lerped
			if (r_lerpmove.value && (ent->lerpflags & LERP_MOVESTEP))
					f = 1;
			//johnfitz

		// interpolate the origin and angles
			for (j=0 ; j<3 ; j++)
			{
				ent->origin[j] = ent->msg_origins[1][j] + f*delta[j];

				d = ent->msg_angles[0][j] - ent->msg_angles[1][j];
				if (d > 180)
					d -= 360;
				else if (d < -180)
					d += 360;
				ent->angles[j] = ent->msg_angles[1][j] + f*d;
			}

		}

#ifdef SUPPORTS_GHOSTING_VIA_DEMOS // Baker change
		if (cls.ghostdemo)
			goto ghost_bypass; // Don't do anything with particles/dlights/etc
#endif // Baker change +

// rotate binary objects locally
		if (ent->model->flags & EF_ROTATE)
			ent->angles[1] = bobjrotate;

		if (ent->effects & EF_BRIGHTFIELD)
			R_EntityParticles (ent);

		if (ent->effects & EF_MUZZLEFLASH)
		{
			vec3_t		fv, rv, uv;

			dl = CL_AllocDlight (i);
			VectorCopy (ent->origin,  dl->origin);
			dl->origin[2] += 16;
			AngleVectors (ent->angles, fv, rv, uv);

			VectorMA (dl->origin, 18, fv, dl->origin);
			dl->radius = 200 + (rand()&31);
			dl->minlight = 32;
			dl->die = cl.time + 0.1;

			//johnfitz -- assume muzzle flash accompanied by muzzle flare, which looks bad when lerped
			if (r_lerpmodels.value != 2)
			{
			if (ent == &cl_entities[cl.viewentity])
				cl.viewent.lerpflags |= LERP_RESETANIM|LERP_RESETANIM2; //no lerping for two frames
			else
				ent->lerpflags |= LERP_RESETANIM|LERP_RESETANIM2; //no lerping for two frames
			}
			//johnfitz
		}
		if (ent->effects & EF_BRIGHTLIGHT)
		{
			dl = CL_AllocDlight (i);
			VectorCopy (ent->origin,  dl->origin);
			dl->origin[2] += 16;
			dl->radius = 400 + (rand()&31);
			dl->die = cl.time + 0.001;
		}
		if (ent->effects & EF_DIMLIGHT)
		{
			dl = CL_AllocDlight (i);
			VectorCopy (ent->origin,  dl->origin);
			dl->radius = 200 + (rand()&31);
			dl->die = cl.time + 0.001;
		}

		if (ent->model->flags & EF_GIB)
			R_RocketTrail (oldorg, ent->origin, 2);
		else if (ent->model->flags & EF_ZOMGIB)
			R_RocketTrail (oldorg, ent->origin, 4);
		else if (ent->model->flags & EF_TRACER)
			R_RocketTrail (oldorg, ent->origin, 3);
		else if (ent->model->flags & EF_TRACER2)
			R_RocketTrail (oldorg, ent->origin, 5);
		else if (ent->model->flags & EF_ROCKET)
		{
			R_RocketTrail (oldorg, ent->origin, 0);
			dl = CL_AllocDlight (i);
			VectorCopy (ent->origin, dl->origin);
			dl->radius = 200;
			dl->die = cl.time + 0.01;
		}
		else if (ent->model->flags & EF_GRENADE)
			R_RocketTrail (oldorg, ent->origin, 1);
		else if (ent->model->flags & EF_TRACER3)
			R_RocketTrail (oldorg, ent->origin, 6);

#ifdef SUPPORTS_GHOSTING_VIA_DEMOS // Baker change
ghost_bypass:
#endif // Baker change +

		ent->forcelink = false;

#ifdef SUPPORTS_GHOSTING_VIA_DEMOS // Baker change
		if (cls.ghostdemo /*making*/ && i == cl.viewentity)
		{
			int i;
			double basetime = cl.mtime[1]; // 0 is newest
			double timeslice = (cl.mtime[0] - cl.mtime[1]);

			vec3_t baseorg  = {ent->msg_origins[1][0], ent->msg_origins[1][1], ent->msg_origins[1][2]};
			vec3_t movedist = { ent->msg_origins[0][0] - ent->msg_origins[1][0], 
								ent->msg_origins[0][1] - ent->msg_origins[1][1], 
								ent->msg_origins[0][2] - ent->msg_origins[1][2]
								};
			vec3_t angle_delta;
			VectorSubtract (ent->msg_angles[0], ent->msg_angles[1], angle_delta);

			for (i = 0; i < 3; i++)
				if (angle_delta[i] > 180)
					angle_delta[i] -= 360;
				else if (angle_delta[i] < -180)
					angle_delta[i] += 360;
			
			for (i = 1; i <=10; i++)
			{
				double		pct = (i/10.0);
				double		fracus = basetime + pct * timeslice ;
				vec3_t		curorg = {baseorg[0] + movedist[0] * pct,
										baseorg[1] + movedist[1] * pct,
										baseorg[2] + movedist[2] * pct
				};
				vec3_t		curang = {ent->msg_angles[1][0] + angle_delta[0] * pct,
										ent->msg_angles[1][1] + angle_delta[1] * pct,
										ent->msg_angles[1][2] + angle_delta[2] * pct
				};
				// Convert everything to a "fixed-point integer for storage"
				int			new_origin[3]	= {curorg[0] * 1000.f, curorg[1] * 1000.f, curorg[2] * 1000.f};
				int			new_angles[3]	= {curang[0] * 1000.f, curang[1] * 1000.f, curang[2] * 1000.f};
	// MAYBE WE NEED TO WRITE THE LERPED STUFF?			
				qboolean	framechange		= (ent->frame != cls.ghost_write.player_frame);
				qboolean	moved			= true; // Until we know otherwise ...
				qboolean	looked			= true; // Until we know otherwise ...

				if (new_origin[0] == cls.ghost_write.player_origin[0])
					if (new_origin[1] == cls.ghost_write.player_origin[1])
						if (new_origin[2] == cls.ghost_write.player_origin[2])
							moved = false;

				if (new_angles[0] == cls.ghost_write.player_angles[PITCH])
					if (new_angles[1] == cls.ghost_write.player_angles[YAW])
						if (new_angles[2] == cls.ghost_write.player_angles[ROLL])
							looked = false;
				
				if (1 || moved || looked || framechange) // Don't worry, a first write will write these as they can't all be 0 (especially player angles).
				{
					// Record type 'P' means player position
					byte	rectype = 'P';
					int		byteswritten = 0;
					int		milliseconds = fracus * 1000.0;
					short	framenum = ent->frame;
					
					cls.ghost_write.player_origin[0] = new_origin[0];
					cls.ghost_write.player_origin[1] = new_origin[1];
					cls.ghost_write.player_origin[2] = new_origin[2];
					cls.ghost_write.player_angles[0] = new_angles[0];
					cls.ghost_write.player_angles[1] = new_angles[1];
					cls.ghost_write.player_angles[2] = new_angles[2];
					
					byteswritten += fwrite (&rectype, 1, sizeof(rectype), cls.ghostfile);			
					byteswritten += fwrite (&framenum, 1, sizeof(framenum), cls.ghostfile);			
					byteswritten += fwrite (&milliseconds, 1, sizeof(milliseconds), cls.ghostfile);
					byteswritten += fwrite (&cls.ghost_write.player_origin[0], 1, sizeof(cls.ghost_write.player_origin[0]), cls.ghostfile);
					byteswritten += fwrite (&cls.ghost_write.player_origin[1], 1, sizeof(cls.ghost_write.player_origin[1]), cls.ghostfile);
					byteswritten += fwrite (&cls.ghost_write.player_origin[2], 1, sizeof(cls.ghost_write.player_origin[2]), cls.ghostfile);
					byteswritten += fwrite (&cls.ghost_write.player_angles[0], 1, sizeof(cls.ghost_write.player_angles[0]), cls.ghostfile);
					byteswritten += fwrite (&cls.ghost_write.player_angles[1], 1, sizeof(cls.ghost_write.player_angles[1]), cls.ghostfile);
					byteswritten += fwrite (&cls.ghost_write.player_angles[2], 1, sizeof(cls.ghost_write.player_angles[2]), cls.ghostfile);
					
					rectype = '\n'; // Newline character ... Makes it easier in textpad to estimate number of records
					byteswritten += fwrite (&rectype, 1, sizeof(rectype), cls.ghostfile); // Padding
					
					if (byteswritten != 32)
						byteswritten = byteswritten;
				}
			}
			continue; // cls.ghostdemo doesn't render so everything below is irrelevant
		}
#endif // Baker change +

		if (i == cl.viewentity && !chase_active.value)
			continue;

		if (cl_numvisedicts < MAX_VISEDICTS)
		{
			cl_visedicts[cl_numvisedicts] = ent;
			cl_numvisedicts++;
#ifdef SUPPORTS_COLORMAPPING_EVERYTHING // Baker change
			if (ent->coloredskin == NULL || SkinTextureChanged (ent))
				ent->coloredskin = R_TranslateNewModelSkinColormap (ent);
#endif // Baker change +
		}
	}
}


/*
===============
CL_ReadFromServer

Read all incoming data from the server
===============
*/
int CL_ReadFromServer (void)
{
	int			ret;
	extern int	num_temp_entities; //johnfitz
	int			num_beams = 0; //johnfitz
	int			num_dlights = 0; //johnfitz
	beam_t		*b; //johnfitz
	dlight_t	*l; //johnfitz
	int			i; //johnfitz


	cl.oldtime = cl.time;
	cl.time += host_frametime;
#ifdef SUPPORTS_DEMO_REWIND // Baker change
	if (!cls.demorewind || !cls.demoplayback)	// by joe
		cl.ctime += host_frametime;
	else
		cl.ctime -= host_frametime;
#endif // Baker change +

	do
	{
		ret = CL_GetMessage ();
		if (ret == -1)
			Host_Error ("CL_ReadFromServer: lost server connection");
		if (!ret)
			break;

		cl.last_received_message = realtime;
		CL_ParseServerMessage ();
	} while (ret && cls.state == ca_connected);

	if (cl_shownet.value)
		Con_Printf ("\n");

	CL_RelinkEntities ();
#ifdef SUPPORTS_GHOSTING_VIA_DEMOS // Baker change
	CL_Ghost_Think ();
#endif // Baker change +
	CL_UpdateTEnts ();

//johnfitz -- devstats

	//visedicts
	if (cl_numvisedicts > 256 && dev_peakstats.visedicts <= 256)
		Con_Warning ("%i visedicts exceeds standard limit of 256.\n", cl_numvisedicts);
	dev_stats.visedicts = cl_numvisedicts;
	dev_peakstats.visedicts = max(cl_numvisedicts, dev_peakstats.visedicts);

	//temp entities
	if (num_temp_entities > 64 && dev_peakstats.tempents <= 64)
		Con_Warning ("%i tempentities exceeds standard limit of 64.\n", num_temp_entities);
	dev_stats.tempents = num_temp_entities;
	dev_peakstats.tempents = max(num_temp_entities, dev_peakstats.tempents);

	//beams
	for (i=0, b=cl_beams ; i< MAX_BEAMS ; i++, b++)
		if (b->model && b->endtime >= cl.time)
			num_beams++;
	if (num_beams > 24 && dev_peakstats.beams <= 24)
		Con_Warning ("%i beams exceeded standard limit of 24.\n", num_beams);
	dev_stats.beams = num_beams;
	dev_peakstats.beams = max(num_beams, dev_peakstats.beams);

	//dlights
	for (i=0, l=cl_dlights ; i<MAX_DLIGHTS ; i++, l++)
		if (l->die >= cl.time && l->radius)
			num_dlights++;
	if (num_dlights > 32 && dev_peakstats.dlights <= 32)
		Con_Warning ("%i dlights exceeded standard limit of 32.\n", num_dlights);
	dev_stats.dlights = num_dlights;
	dev_peakstats.dlights = max(num_dlights, dev_peakstats.dlights);

//johnfitz

#ifdef SUPPORTS_AVI_CAPTURE // Baker change
#ifdef SUPPORTS_DEMO_PROGRESS_ESTIMATION // Baker change
	if (cls.demoplayback && cls.capturedemo /*cls.demonum == -1 && !cls.timedemo && !cls.titledemo*/)
	{
		static float olddrealtime; // Yay.  Another timer.  Sheesh.
		float timeslice = realtime - olddrealtime;
		olddrealtime = realtime;

		if (cl.paused & 2)
			timeslice = 0;

		// If we have no start cltime, fill it in now
		if (!cls.demo_cltime_start)
		{
			cls.demo_cltime_start = cl.time;
			cls.demo_cltime_elapsed = 0;
		}
		else cls.demo_cltime_elapsed += host_frametime;

		// If we have no start hosttime, fill it in now
		if (!cls.demo_hosttime_start)
		{
			cls.demo_hosttime_start = realtime;
			cls.demo_hosttime_elapsed = 0;	
		}
		else cls.demo_hosttime_elapsed += timeslice; // Advance time only if we are not paused

		while (1)
		{
			// This is the "percentage" (0 to 1) of the demoplay that has been completed.
			float completed_amount = (cls.demo_offset_current - cls.demo_offset_start)/(float)cls.demo_file_length;
			float remaining_time = 0;
			int minutes, seconds;
			char tempstring[256];
			extern int vid_default;
			extern char movie_codec[12];

			if (vid_default != 0)
				break; // Don't bother, we are in fullscreen mode.

			if (timeslice = 0)
				break; // Don't bother updating the caption if we are paused

			if (cls.demo_hosttime_elapsed)
				remaining_time = (cls.demo_hosttime_elapsed / completed_amount) - cls.demo_hosttime_elapsed;

			minutes = COM_Minutes((int)remaining_time);
			seconds = COM_Seconds((int)remaining_time);

			sprintf (tempstring, "Demo: %s (%3.1f%% elapsed: %4.1f secs) - Estimated Remaining %d:%02d (Capturing: %s)", cls.demoname, completed_amount * 100, cls.demo_hosttime_elapsed, (int)minutes, (int)seconds, movie_codec);

			Sys_SetWindowCaption (tempstring);
			break;
		}

	}
#endif // Baker change +
#endif // Baker change +
//
// bring the links up to date
//
	return 0;
}

/*
=================
CL_SendCmd
=================
*/
void CL_SendCmd (void)
{
	usercmd_t		cmd;

	if (cls.state != ca_connected)
		return;

	if (cls.signon == SIGNONS)
	{
	// get basic movement from keyboard
		CL_BaseMove (&cmd);

	// allow mice or other external controllers to add to the move
		IN_Move (&cmd);

	// send the unreliable message
		CL_SendMove (&cmd);

	}

	if (cls.demoplayback)
	{
		SZ_Clear (&cls.message);
		return;
	}

// send the reliable message
	if (!cls.message.cursize)
		return;		// no message at all

	if (!NET_CanSendMessage (cls.netcon))
	{
		Con_DPrintf ("CL_WriteToServer: can't send\n");
		return;
	}

	if (NET_SendMessage (cls.netcon, &cls.message) == -1)
		Host_Error ("CL_WriteToServer: lost server connection");

	SZ_Clear (&cls.message);
}

/*
=============
CL_Tracepos_f -- johnfitz

display impact point of trace along VPN
=============
*/
void CL_Tracepos_f (void)
{
	vec3_t	v, w;

	VectorScale(vpn, 8192.0, v);
	TraceLine(r_refdef.vieworg, v, w);

	if (Length(w) == 0)
		Con_Printf ("Tracepos: trace didn't hit anything\n");
	else
		Con_Printf ("Tracepos: (%i %i %i)\n", (int)w[0], (int)w[1], (int)w[2]);
}

/*
=============
CL_Viewpos_f -- johnfitz

display client's position and angles
=============
*/
void CL_Viewpos_f (void)
{
#if 0
	//camera position
	Con_Printf ("Viewpos: (%i %i %i) %i %i %i\n",
		(int)r_refdef.vieworg[0],
		(int)r_refdef.vieworg[1],
		(int)r_refdef.vieworg[2],
		(int)r_refdef.viewangles[PITCH],
		(int)r_refdef.viewangles[YAW],
		(int)r_refdef.viewangles[ROLL]);
#else
	//player position
	Con_Printf ("Viewpos: (%i %i %i) %i %i %i\n",
		(int)cl_entities[cl.viewentity].origin[0],
		(int)cl_entities[cl.viewentity].origin[1],
		(int)cl_entities[cl.viewentity].origin[2],
		(int)cl.viewangles[PITCH],
		(int)cl.viewangles[YAW],
		(int)cl.viewangles[ROLL]);
#endif
}

/*
=================
CL_Init
=================
*/
void CL_Init (void)
{
#ifdef SUPPORTS_TITLE_DEMOS // Baker change
extern cvar_t cl_titledemos_list;
#endif // Baker change +
	SZ_Alloc (&cls.message, 1024);

	CL_InitInput ();
	CL_InitTEnts ();

	Cvar_RegisterVariable (&cl_name, NULL);
	Cvar_RegisterVariable (&cl_color, NULL);
	Cvar_RegisterVariable (&cl_upspeed, NULL);
	Cvar_RegisterVariable (&cl_forwardspeed, NULL);
	Cvar_RegisterVariable (&cl_backspeed, NULL);
	Cvar_RegisterVariable (&cl_sidespeed, NULL);
	Cvar_RegisterVariable (&cl_movespeedkey, NULL);
	Cvar_RegisterVariable (&cl_yawspeed, NULL);
	Cvar_RegisterVariable (&cl_pitchspeed, NULL);
	Cvar_RegisterVariable (&cl_anglespeedkey, NULL);
	Cvar_RegisterVariable (&cl_shownet, NULL);
	Cvar_RegisterVariable (&cl_nolerp, NULL);
	Cvar_RegisterVariable (&lookspring, NULL);
	Cvar_RegisterVariable (&lookstrafe, NULL);
	Cvar_RegisterVariable (&sensitivity, NULL);

	Cvar_RegisterVariable (&m_pitch, NULL);
	Cvar_RegisterVariable (&m_yaw, NULL);
	Cvar_RegisterVariable (&m_forward, NULL);
	Cvar_RegisterVariable (&m_side, NULL);

	Cvar_RegisterVariable (&cl_maxpitch, NULL); //johnfitz -- variable pitch clamping
	Cvar_RegisterVariable (&cl_minpitch, NULL); //johnfitz -- variable pitch clamping
#ifdef SUPPORTS_TITLE_DEMOS // Baker change
	Cvar_RegisterVariable (&cl_titledemos_list, NULL);
#endif // Baker change +

	Cmd_AddCommand ("entities", CL_PrintEntities_f);
	Cmd_AddCommand ("disconnect", CL_Disconnect_f);
	Cmd_AddCommand ("record", CL_Record_f);
	Cmd_AddCommand ("stop", CL_Stop_f);
	Cmd_AddCommand ("playdemo", CL_PlayDemo_f);

#ifdef TOUCHUP_CLEAR_DEMOQUEUE_PLAYDEMO // Baker change
	Cmd_AddCommand ("nextstartdemo", CL_PlayDemo_NextStartDemo_f);
#endif // Baker change +

	Cmd_AddCommand ("timedemo", CL_TimeDemo_f);
#ifdef SUPPORTS_GHOSTING_VIA_DEMOS // Baker change
	Cmd_AddCommand ("ghostdemo", CL_GhostDemo_f);
#endif // Baker change +

	Cmd_AddCommand ("tracepos", CL_Tracepos_f); //johnfitz
	Cmd_AddCommand ("viewpos", CL_Viewpos_f); //johnfitz
}

